from . import tools
from ..random_explorer import RandomExplorer, RandomExplorerParams


class Agent(RandomExplorer, tools.Agent):
    pass


def test_selected_action_is_sampled_more_frequently():
    rep = RandomExplorerParams(init_ep=0.6, final_ep=0.01, decay_ep=1e-4)
    test_agent = Agent(ep=rep)
    actions = list(range(5))
    chosen_actions = {}
    for _ in range(1000):
        a = test_agent.ex_choose(actions, lambda x: x[0])
        if a not in chosen_actions:
            chosen_actions[a] = 1
        else:
            chosen_actions[a] += 1

    for a in chosen_actions:
        assert chosen_actions[a] > 0
        if a == 0:
            continue
        assert chosen_actions[0] > chosen_actions[a]




import typing as T
import gym
import numpy as np

from .environment import Environment


class CartPoleFrames(Environment):
    def __init__(self):
        self.env: gym.Env = gym.make('CartPole-v1').unwrapped
        self.env.reset()

    def get_observation_space(self) -> T.Tuple[int]:
        screen: np.ndarray = self.env.render(mode="rgb_array")
        return screen.shape

    def get_action_space(self) -> T.Tuple[int]:
        return tuple(range(self.env.action_space.n))

    def reset(self) -> np.ndarray:
        self.env.reset()
        return self.env.render(mode="rgb_array").copy()

    def do_step(self, action: int) -> T.Tuple[np.ndarray, float, bool]:
        _, r, f, _ = self.env.step(action)
        return self.env.render(mode="rgb_array").copy(), r, f

    def render(self) -> None:
        self.env.render()

    def close(self) -> None:
        self.env.close()



import typing as T
from abc import ABC, abstractmethod
import random
import numpy as np
import asyncio
import torch

from ..agents import AnyAgent
from ..environments import Environment
from .individual import Individual
from .models import EvolutionProgress, EvolutionerParams, T_EParams, EvolvingFloat, EvolvingBool, EvolvingInt


class Evolutioner(ABC):
    def __init__(self,
                 env: Environment,
                 init_params: T_EParams,
                 evolutioner_params: EvolutionerParams):

        self.env: Environment = env
        self.evolutioner_params: EvolutionerParams = evolutioner_params
        self.init_params: T_EParams = init_params
        self.progress_callback: T.Callable[[EvolutionProgress], None] = lambda x: None
        self.fitness_function: T.Callable[[np.ndarray], float] = lambda x: np.mean(x).item()

    @abstractmethod
    def agent_factory(self, params: T_EParams, state_dict: T.Optional[dict]) -> AnyAgent:
        raise NotImplementedError()

    def set_progress_callback(self, cbk: T.Callable[[EvolutionProgress], None]) -> None:
        self.progress_callback = cbk

    def create_generation(self, params: T_EParams, state_dict: dict) -> T.Tuple[T.List[Individual], T.List[T_EParams]]:
        result: T.List[Individual] = []
        params_result: T.List[T_EParams] = []
        size = self.evolutioner_params.generation_size
        for i in range(size):
            mutated_params = self.mutate(params, i/size)
            agent = self.agent_factory(mutated_params, state_dict)
            result.append(Individual(agent, self.env))
            params_result.append(mutated_params)
        return result, params_result

    @staticmethod
    def mutate(params: T_EParams, scale_mutation: float) -> T_EParams:
        result = {}
        for k, v in params.items():
            if isinstance(v, EvolvingFloat):
                v_dict = v.__dict__
                mutation = (1 if random.random() > 0.5 else -1) * scale_mutation * v.mutation_step * random.random()
                v_dict["value"] += mutation
                result[k] = EvolvingFloat(**v_dict)
            elif isinstance(v, EvolvingInt):
                v_dict = v.__dict__
                mutation = (1 if random.random() > 0.5 else -1) * scale_mutation * v.mutation_step * random.random()
                v_dict["value"] = int(mutation + v_dict["value"])
                result[k] = EvolvingInt(**v_dict)
            elif isinstance(v, EvolvingBool):
                v_dict = v.__dict__
                mutation = scale_mutation * v.mutation_step
                if random.random() > mutation:
                    v_dict["value"] = not v_dict["value"]
                result[k] = EvolvingBool(**v_dict)
            else:
                raise ValueError(f"value from param {k} has type {type(v)}. Only types int, float and bool are valid")
        return result

    def evolve(self, finish_condition: T.Callable[[EvolutionProgress], bool]) -> None:
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._evolve(finish_condition))

    async def _evolve(self, finish_condition: T.Callable[[EvolutionProgress], bool]) -> None:
        torch.multiprocessing.set_start_method("spawn")
        params = self.init_params.copy()
        next_state_dict = None
        generation_count = 0
        while True:
            generation, mutated_params = self.create_generation(params, next_state_dict)
            results: T.List[T.Union[float, None]] = [None for _ in range(len(generation))]
            state_dicts: T.List[T.Union[dict, None]] = [None for _ in range(len(generation))]

            tasks: T.List[T.Coroutine] = []
            completed = []
            for i in range(len(generation)):
                async def task(index: int):
                    life_result, state_dict = await generation[index].parallel_life(index)
                    results[index] = self.fitness_function(life_result)
                    state_dicts[index] = state_dict
                    completed.append(None)
                    print("=============================")
                    print(f"completed {len(completed)}/{len(generation)}: {results[index]}")
                    for k, v in mutated_params[index].items():
                        print(k+":", v.value)

                tasks.append(task(i))
                if i >= len(generation) - 1:
                    await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
                    break
                if len(tasks) >= self.evolutioner_params.workers:
                    _, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
                    tasks = list(pending)

            best_index = np.argmax(results).item()
            progress = EvolutionProgress(
                results,
                mutated_params,
                generation_count,
                best_index
            )

            self.progress_callback(progress)
            if finish_condition(progress):
                break
            generation_count += 1
            params = mutated_params[best_index]
            next_state_dict = state_dicts[best_index]



from abc import ABC, abstractmethod
import numpy as np
import typing as T


class Environment(ABC):
    last_s: T.Union[None, np.ndarray] = None

    @abstractmethod
    def get_observation_space(self) -> T.Tuple[int]:
        raise NotImplementedError()

    @abstractmethod
    def get_action_space(self) -> T.Tuple[int]:
        raise NotImplementedError()

    @abstractmethod
    def reset(self) -> np.ndarray:
        raise NotImplementedError()

    def step(self, action: int) -> T.Tuple[np.ndarray, float, bool]:
        s, r, f = self.do_step(action)
        self.last_s = s
        return s, r, f

    @abstractmethod
    def do_step(self, action: int) -> T.Tuple[np.ndarray, float, bool]:
        raise NotImplementedError()

    @abstractmethod
    def render(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError()



import random
import typing as T
from abc import ABC

from .base.params import ReplayBufferEntry, PrioritizedReplayBufferParams
from .base.replay_buffer import ReplayBuffer
from ..base.models import LearningStep
from ..replay_buffers.base.segment_trees import SumSegmentTree, MinSegmentTree


class PrioritizedReplayBuffer(ReplayBuffer, ABC):
    def __init__(self, rp: PrioritizedReplayBufferParams = PrioritizedReplayBufferParams(), *args, **kwargs):
        if not isinstance(rp, PrioritizedReplayBufferParams):
            raise ValueError("argument rp must be an instance of PrioritizedReplayBufferParams")
        if rp.alpha < 0:
            raise ValueError("alpha must be >= 0")

        self.max_priority: float = 1.0
        self.rp: PrioritizedReplayBufferParams = rp
        self.beta: float = rp.init_beta

        tree_capacity = 1
        while tree_capacity < self.rp.max_len:
            tree_capacity *= 2

        self.sum_tree = SumSegmentTree(tree_capacity)
        self.min_tree = MinSegmentTree(tree_capacity)
        super(PrioritizedReplayBuffer, self).__init__(rp, *args, **kwargs)

    def _increase_beta(self, *_, **__):
        self.log.debug(f"increase beta for {type(self).__name__} triggered")
        if self.beta < self.rp.final_beta:
            self.beta += self.rp.increase_beta
        elif self.beta > self.rp.final_beta:
            self.beta = self.rp.final_beta

    def rp_add(self, entry: T.Union[ReplayBufferEntry]) -> bool:
        self.sum_tree[self.ptr] = self.max_priority ** self.rp.alpha
        self.min_tree[self.ptr] = self.max_priority ** self.rp.alpha
        return super(PrioritizedReplayBuffer, self).rp_add(entry)

    def rp_sample(self, batch_size: int) -> T.List[ReplayBufferEntry]:
        indices = self._sample_proportional(batch_size)

        entries: T.List[ReplayBufferEntry] = [self.records[i] for i in indices]
        for entry in entries:
            entry.weight = self._calculate_weight(entry.index)

        return entries

    def _refactor_priorities(self, indices: T.List[int], priorities: T.List[float]):
        if len(indices) != len(priorities):
            raise ValueError("indices and priorities must have the same length")

        for idx, priority in zip(indices, priorities):
            if priority <= 0:
                raise ValueError("priority for index "+str(idx)+" is less than 0")
            if not 0 <= idx < self.rp_get_length():
                raise ValueError("index "+str(idx)+" out of range")

            self.sum_tree[idx] = priority ** self.rp.alpha
            self.min_tree[idx] = priority ** self.rp.alpha

            self.max_priority = max(self.max_priority, priority)

    def _sample_proportional(self, batch_size: int) -> T.List[int]:
        indices = []
        p_total = self.sum_tree.sum(0, self.rp_get_length())
        segment = p_total / batch_size

        for i in range(batch_size):
            a = segment * i
            b = segment * (i + 1)
            upperbound = random.uniform(a, b)
            idx = self.sum_tree.retrieve(upperbound)
            indices.append(idx)

        return indices

    def _calculate_weight(self, idx: int):
        p_min = self.min_tree.min() / self.sum_tree.sum()
        max_weight = (p_min * self.rp_get_length()) ** (-self.beta)

        p_sample = self.sum_tree[idx] / self.sum_tree.sum()
        weight = (p_sample * self.rp_get_length()) ** (-self.beta)
        weight = weight / max_weight

        return weight

    def _update_priorities(self, learning_step: LearningStep) -> None:
        self.log.debug(f"update priorities for {type(self).__name__} triggered")
        self._refactor_priorities(
            [e.index for e in learning_step.batch],
            [abs(x - y) + 1e-7 for x, y in zip(learning_step.x, learning_step.y)]
        )

    def rp_clear(self):
        super(PrioritizedReplayBuffer, self).rp_clear()
        self.sum_tree.clear()
        self.min_tree.clear()

    def rp_link(self):
        self.log.info(f"linking {type(self).__name__} priority...")
        self.add_step_callback("prioritized replay buffer increase beta", self._increase_beta)
        self.add_learn_callback("prioritized replay buffer update priorities", self._update_priorities)

    def rp_get_stats(self) -> T.Dict[str, float]:
        stats = super(PrioritizedReplayBuffer, self).rp_get_stats()
        stats.update({"Prioritized Replay Buffer Beta": self.beta})
        return stats



import time
import subprocess
import webbrowser
import threading
from torch.utils.tensorboard import SummaryWriter


class TensorBoard(SummaryWriter):
    def __init__(self, path: str):
        super(TensorBoard, self).__init__(path)
        self.process = None
        threading.Thread(target=lambda: time.sleep(3) or self.launch(path)).start()

    def launch(self, path):
        try:
            self.process = subprocess.Popen(["tensorboard", "--logdir=" + path, "--load_fast=false"], shell=False)
            threading.Thread(target=lambda: time.sleep(1) or webbrowser.open("http://localhost:6006")).start()
        except FileNotFoundError:
            print("tensorboard is not installed")

    def __del__(self):
        if self.process:
            self.process.kill()



from .base.n_steps_replay_buffer import NStepsReplayBuffer
from .random_replay_buffer import RandomReplayBuffer
from .base.params import NStepRandomReplayBufferParams
from abc import ABC


class NStepsRandomReplayBuffer(NStepsReplayBuffer, RandomReplayBuffer, ABC):
    def __init__(self, rp: NStepRandomReplayBufferParams = NStepRandomReplayBufferParams(), *args, **kwargs):
        if not isinstance(rp, NStepRandomReplayBufferParams):
            raise ValueError("argument rp must be an instance of NStepPrioritizedReplayBufferParams")
        super(NStepsRandomReplayBuffer, self).__init__(rp, *args, **kwargs)

    def rp_link(self):
        NStepsReplayBuffer.rp_link(self)
        RandomReplayBuffer.rp_link(self)



from . import tools
from .test_n_steps_random_replay_buffer import is_n_step
from .test_prioritized_replay_buffer import is_prioritized
from ...replay_buffers import NStepPrioritizedReplayBufferParams, NStepsPrioritizedReplayBuffer


class Agent(NStepsPrioritizedReplayBuffer, tools.Agent):
    pass


def test_elements_are_n_step():
    nprp = NStepPrioritizedReplayBufferParams(max_len=15, n_step=3, alpha=0.5, init_beta=0.9, final_beta=1.0, increase_beta=1e-5)
    test_agent = Agent(rp=nprp)
    is_n_step(test_agent)


def test_elements_retrieved_are_prioritized():
    nprp = NStepPrioritizedReplayBufferParams(max_len=10, n_step=1, alpha=0.5, init_beta=0.9, final_beta=1.0, increase_beta=1e-5)
    test_agent = Agent(rp=nprp)
    is_prioritized(test_agent)



import typing as T
from abc import ABC

import torch

from . import models, base_object
from .. import interfaces
from ...environments import Environment


class BaseAgent(base_object.BaseObject,
                interfaces.AgentEventInterface,
                interfaces.ExplorerInterface,
                interfaces.ReplayBufferInterface,
                ABC):
    def __init__(self, action_space: int, hp: models.HyperParams, use_gpu: bool = True):
        super(BaseAgent, self).__init__()
        self.device: str = "cpu"
        if use_gpu:
            if not torch.cuda.is_available():
                self.log.warning("cuda is not available, CPU will be used")
            else:
                self.device = "cuda"
        self.action_space = action_space
        self.hyper_params = hp
        self.step_callbacks: T.Dict[str, models.TStepCallback] = {}
        self.progress_callbacks: T.Dict[str, models.TProgressCallback] = {}
        self.learning_callbacks: T.Dict[str, models.TLearnCallback] = {}
        self.accumulate_rewards: bool = True
        self.rp_link()
        self.ex_link()

    def get_self_class_name(self):
        return self.__class__.__name__

    @staticmethod
    def last_layer_factory(in_features: int, out_features: int) -> torch.nn.Module:
        return torch.nn.Linear(in_features, out_features)

    def health_check(self, env: Environment) -> None:
        self.log.info("checking the model is healthy...")
        s = env.reset()
        self.log.debug(f"state for testing health is:\n{s}")
        self.log.info("testing preprocessing...")
        try:
            self.preprocess(s)
        except Exception as e:
            self.log.error("error while testing preprocessing")
            raise e
        self.log.info("preprocessing is correct")
        self.log.info("testing inference...")
        try:
            self.infer(s)
        except Exception as e:
            self.log.error("error while testing inference")
            raise e
        self.log.info("inference is correct")
        self.log.info("testing learning...")
        while self.rp_get_length() < 2:
            a = 0
            s_, r, final = env.step(a)
            self.rp_add(models.ReplayBufferEntry(s, s_, a, r, final))
            s = s_
            if final:
                s = env.reset()

        batch = self.rp_sample(2)
        try:
            self.learn(batch)
        except Exception as e:
            self.log.error("error while testing learning")
            raise e

        self.log.info("learning is correct")
        self.log.info("model is healthy!")
        self.rp_clear()

    def add_step_callback(self, label: str, cbk: models.TStepCallback) -> None:
        if label in self.step_callbacks:
            self.log.warning(f"overwriting step callback with label {label}")
        self.step_callbacks[label] = cbk
        self.log.debug(f"added step callback {label}: {cbk}")

    def add_progress_callback(self, label: str, cbk: models.TProgressCallback) -> None:
        if label in self.progress_callbacks:
            self.log.warning(f"overwriting progress callback with label {label}")
        self.progress_callbacks[label] = cbk
        self.log.debug(f"added progress callback {label}: {cbk}")

    def add_learn_callback(self, label: str, cbk: models.TLearnCallback) -> None:
        if label in self.learning_callbacks:
            self.log.warning(f"overwriting learning callback with label {label}")
        self.learning_callbacks[label] = cbk
        self.log.debug(f"added learning callback {label}: {cbk}")

    def call_step_callbacks(self, training_step: models.TrainingStep) -> None:
        self.log.debug("calling step callbacks...")
        for label, cbk in self.step_callbacks.items():
            self.log.debug(f"calling step callback {label}, {cbk}")
            cbk(training_step)
        self.log.debug("all step callbacks called")

    def call_progress_callbacks(self, training_progress: models.TrainingProgress) -> bool:
        self.log.debug("calling progress callbacks...")
        must_exit = False
        for label, cbk in self.progress_callbacks.items():
            self.log.debug(f"calling progress callback {label}, {cbk}")
            may_exit = cbk(training_progress)
            if may_exit:
                must_exit = True
                self.log.warning(f"progress callback {label} said that training should end")
        self.log.debug("all progress callbacks called")
        return must_exit

    def call_learn_callbacks(self, learning_step: models.LearningStep) -> None:
        self.log.debug("calling learning callbacks...")
        for label, cbk in self.learning_callbacks.items():
            self.log.debug(f"calling learning callback {label}, {cbk}")
            cbk(learning_step)
        self.log.debug("all learning callbacks called")

    def forward_hook(self, module: torch.nn.Module, x: T.Tuple[torch.Tensor], y: torch.Tensor) -> None:
        pass

    def build_model(self) -> torch.nn.Module:
        self.log.info("building model from model factory...")
        model = self.model_factory()
        model.register_forward_hook(self.forward_hook)
        self.log.info("model built correctly, applying last layers modifier...")
        model = self.last_layers_model_modifier(model)
        self.log.info("last layers modified correctly, applying agent specification")
        model = self.agent_specification_model_modifier(model)
        self.log.info("agent specification applied correctly")
        return model

    def preprocess(self, x: T.Iterable) -> torch.Tensor:
        return torch.unsqueeze(torch.tensor(x, dtype=torch.float32), 0)



import operator

from .segment_tree import SegmentTree


class SumSegmentTree(SegmentTree):
    def __init__(self, capacity: int):
        super(SumSegmentTree, self).__init__(capacity=capacity, operation=operator.add, init_value=0.0)

    def sum(self, start: int = 0, end: int = 0) -> float:
        return super(SumSegmentTree, self).operate(start, end)

    def retrieve(self, upperbound: float) -> int:
        """Find the highest index `i` about upper bound in the tree"""
        # TODO: Check assert case and fix bug
        assert 0 <= upperbound <= self.sum() + 1e-5, "upperbound: {}".format(upperbound)

        idx = 1

        while idx < self.capacity:  # while non-leaf
            left = 2 * idx
            right = left + 1
            if self.tree[left] > upperbound:
                idx = 2 * idx
            else:
                upperbound -= self.tree[left]
                idx = right
        return idx - self.capacity





import asyncio
import json
import multiprocessing
import os
import pickle
import time
import typing as T

from ..agents import AnyAgent, TrainingProgress
from ..environments import Environment


class Individual:
    def __init__(self, agent: AnyAgent, env: Environment):
        self.agent: AnyAgent = agent
        self.env: Environment = env
        self.history: T.List[float] = []

    def _on_progress(self, progress: TrainingProgress) -> bool:
        self.history.append(progress.total_reward)
        return False

    def life(self, dump_file: str = "", state_dump_file: str = "") -> T.List[float]:
        self.agent.add_progress_callback("evolution callback", self._on_progress)
        self.agent.tensor_board_log = False
        self.agent.save_progress = False
        self.agent.train(self.env)
        self.env.close()
        if dump_file != "":
            json.dump(self.history, open(dump_file, "w"))
        if state_dump_file != "":
            pickle.dump(self.agent.get_state_dict(), open(state_dump_file, "wb"))
        return self.history

    async def parallel_life(self, worker_id: int, timeout: int = 0) -> T.Tuple[T.List[float], dict]:
        now = str(time.time())
        file_name = f"{worker_id}_result_{now}.json"
        state_file_name = f"{worker_id}_state_dict_{now}.json"

        process = multiprocessing.Process(target=self.life, args=(file_name, state_file_name))
        process.start()
        start = time.time()
        while True:
            if timeout and time.time()-start > timeout:
                process.kill()
                raise TimeoutError("worker "+str(worker_id)+" timeout")

            if os.path.isfile(file_name) and os.path.isfile(state_file_name):
                await asyncio.sleep(1)
                result = json.load(open(file_name))
                state_dict = pickle.load(open(state_file_name, "rb"))
                os.remove(file_name)
                os.remove(state_file_name)
                return result, state_dict
            elif not process.is_alive():
                raise RuntimeError("worker "+str(worker_id)+" died without returning a result")

            await asyncio.sleep(1)



import typing as T


class SegmentTree:
    def __init__(self, capacity: int, operation: T.Callable, init_value: float):
        """Initialization.
        Args:
            capacity (int)
            operation (function)
            init_value (float)
        """
        assert capacity > 0 and capacity & (capacity - 1) == 0, "capacity must be positive and a power of 2."
        self.init_value = init_value
        self.capacity = capacity
        self.tree = [init_value for _ in range(2 * capacity)]
        self.operation = operation

    def clear(self):
        self.tree = [self.init_value for _ in range(2 * self.capacity)]

    def _operate_helper(self, start: int, end: int, node: int, node_start: int, node_end: int) -> float:
        """Returns result of operation in segment."""
        if start == node_start and end == node_end:
            return self.tree[node]
        mid = (node_start + node_end) // 2
        if end <= mid:
            return self._operate_helper(start, end, 2 * node, node_start, mid)
        else:
            if mid + 1 <= start:
                return self._operate_helper(start, end, 2 * node + 1, mid + 1, node_end)
            else:
                return self.operation(
                    self._operate_helper(start, mid, 2 * node, node_start, mid),
                    self._operate_helper(mid + 1, end, 2 * node + 1, mid + 1, node_end),
                )

    def operate(self, start: int = 0, end: int = 0) -> float:
        """Returns result of applying `self.operation`."""
        if end <= 0:
            end += self.capacity
        end -= 1

        return self._operate_helper(start, end, 1, 0, self.capacity - 1)

    def __setitem__(self, idx: int, val: float):
        """Set value in tree."""
        idx += self.capacity
        self.tree[idx] = val

        idx //= 2
        while idx >= 1:
            self.tree[idx] = self.operation(self.tree[2 * idx], self.tree[2 * idx + 1])
            idx //= 2

    def __getitem__(self, idx: int) -> float:
        """Get real value in leaf node of tree."""
        assert 0 <= idx < self.capacity

        return self.tree[self.capacity + idx]



from .base.params import NStepPrioritizedReplayBufferParams
from .base.n_steps_replay_buffer import NStepsReplayBuffer
from .priorized_replay_buffer import PrioritizedReplayBuffer
from abc import ABC


class NStepsPrioritizedReplayBuffer(PrioritizedReplayBuffer, NStepsReplayBuffer, ABC):
    def __init__(self, rp: NStepPrioritizedReplayBufferParams = NStepPrioritizedReplayBufferParams(), *args, **kwargs):
        if not isinstance(rp, NStepPrioritizedReplayBufferParams):
            raise ValueError("argument rp must be an instance of NStepPrioritizedReplayBufferParams")
        super(NStepsPrioritizedReplayBuffer, self).__init__(rp, *args, **kwargs)

    def rp_link(self):
        PrioritizedReplayBuffer.rp_link(self)
        NStepsReplayBuffer.rp_link(self)



import typing as T
from abc import ABC
from random import random, randrange

import torch

from ..explorers.base.explorer import Explorer
from ..explorers.base.params import RandomExplorerParams


class RandomExplorer(Explorer, ABC):
    def __init__(self, ep: RandomExplorerParams = RandomExplorerParams(), *args, **kwargs):
        if not isinstance(ep, RandomExplorerParams):
            raise ValueError("argument ep must be an instance of RandomExplorerParams")
        if not 0 <= ep.init_ep <= 1:
            raise ValueError("initial epsilon must be between 0 and 1")
        elif not 0 <= ep.final_ep <= 1:
            raise ValueError("final epsilon must be between 0 and 1")
        elif ep.final_ep >= ep.init_ep:
            raise ValueError("final epsilon must be less than initial epsilon")
        elif not 0 < ep.decay_ep < 1:
            raise ValueError("decay epsilon must be between 0 and 1")
        self.ep: RandomExplorerParams = ep
        self.epsilon: float = ep.init_ep
        self.arrived_to_minimum: bool = False
        super(RandomExplorer, self).__init__(*args, **kwargs)

    def last_layers_model_modifier(self, model: torch.nn.Module) -> torch.nn.Module:
        return model

    def decay(self, *_, **__) -> None:
        self.log.debug(f"decay epsilon for {type(self).__name__} triggered")
        if self.epsilon > self.ep.final_ep:
            self.epsilon -= self.ep.decay_ep
        elif self.epsilon < self.ep.final_ep:
            self.epsilon = self.ep.final_ep
        elif not self.arrived_to_minimum:
            self.log.info("epsilon arrived to minimum")
            self.arrived_to_minimum = True

    def ex_choose(self, actions: T.List[float], f: T.Callable[[T.List[float]], int]) -> int:
        if random() > self.epsilon:
            return f(actions)
        else:
            return randrange(0, len(actions))

    def ex_link(self):
        self.add_step_callback("random explorer decay", self.decay)

    def ex_get_stats(self) -> T.Dict[str, float]:
        return {"Random Explorer Epsilon": self.epsilon}



import typing as T
from abc import ABC

import torch

from .base.models import DuelingDqnHyperParams
from .dqn import DqnAgent
from .explorers.noisy_explorer import NoisyLinear


class DuelingDqnNetwork(torch.nn.Module):
    def __init__(self,
                 model: torch.nn.Module,
                 action_space: int,
                 last_layer_factory: T.Callable[[int, int], torch.nn.Module]):
        super(DuelingDqnNetwork, self).__init__()
        self.model = model
        last_layer = list(model.modules())[-1]
        if not isinstance(last_layer, (torch.nn.Linear, NoisyLinear)):
            raise ValueError("the model you have created must have a torch.nn.Linear or "
                             "agents.explorers.noisy_explorer.NoisyLinear in the last layer")

        if last_layer.out_features == action_space:
            print("WARNING: detected same number of features in the output of the model than the action space")

        self.value = last_layer_factory(last_layer.out_features, 1)
        self.advantage = last_layer_factory(last_layer.out_features, action_space)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)
        v = self.value(x)
        a = self.advantage(x)
        return v + a - a.mean(dim=-1, keepdim=True)


class DuelingDqnAgent(DqnAgent, ABC):
    def __init__(self,
                 action_space: int,
                 hp: DuelingDqnHyperParams = DuelingDqnHyperParams(),
                 use_gpu: bool = True):
        super(DuelingDqnAgent, self).__init__(action_space, hp, use_gpu)

    def dqn_model_wrapper(self, model: torch.nn.Module) -> torch.nn.Module:
        return DuelingDqnNetwork(model, self.action_space, self.last_layer_factory)
